/* -*- Mode: C; c-basic-offset: 2; indent-tabs-mode: nil -*-
 *
 * Pigment media rendering library
 *
 * Copyright © 2006, 2007, 2008 Fluendo Embedded S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Author(s): Loïc Molinari <loic@fluendo.com>
 *            Julien Moutte <julien@fluendo.com>
 */

/**
 * SECTION:pgmdrawable
 * @short_description: A base class used for embedding media types.
 * @see_also: #PgmImage, #PgmText, #PgmCanvas.
 *
 * <refsect2>
 * <para>
 * #PgmDrawable represents a visual object on which you can embed media types
 * such as text, images and video, or 2D graphics. It is meant to be a base
 * abstract class for more specific visual objects such as #PgmImage or
 * #PgmText. Drawables are not visible until they are added to a #PgmCanvas.
 * A drawable never knows about physical on screen coordinates such as pixels,
 * and use floating point canvas coordinates. Drawables are positioned in 3D
 * and then projected on a #PgmViewport.
 * </para>
 * <title>Manipulating drawables</title>
 * <para>
 * The base class implements simple drawing methods that are common to the
 * different subclasses. For example if you want to draw a left aligned green
 * text you would do something like that:
 * <example id="pigment-drawable-functions">
 * <title>Calling drawing functions on a drawable</title>
 * <programlisting language="c">
 * pgm_drawable_set_fg_color (text, 0, 255, 0, 255);
 * pgm_drawable_set_bg_color (text, 255, 0, 0, 255);
 * pgm_text_set_label (PGM_TEXT (text), "Hello world!");
 * </programlisting>
 * </example>
 * You are basically using a mix of drawable calls and text calls on the
 * same object at different level of the object hierarchy.
 * </para>
 * <title>Signal connections</title>
 * <para>
 * The <link linkend="PgmDrawable-changed">changed</link> signal is fired
 * whenever a property of a drawable is changed. There are also signals for
 * handling mouse picking. For instance, you just need to connect a callback to
 * the <link linkend="PgmDrawable-changed">clicked</link> signal to be notified
 * of a click on a drawable.
 * </para>
 * </refsect2>
 *
 * Last reviewed on 2007-09-25 (0.3.1)
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <string.h>
#include "pgmdrawable.h"
#include "pgmcanvas.h"
#include "pgmmarshal.h"
#include "pgmenumtypes.h"

/* Double-click delay in milliseconds */
#define DOUBLE_CLICK_DELAY 250

GST_DEBUG_CATEGORY_STATIC (pgm_drawable_debug);
#define GST_CAT_DEFAULT pgm_drawable_debug

/* Drawable signals */
enum {
  CHANGED,
  PRESSED,
  PRESSURED,
  RELEASED,
  CLICKED,
  DOUBLE_CLICKED,
  SCROLLED,
  DRAG_BEGIN,
  DRAG_MOTION,
  DRAG_END,
  MOTION,
  ENTERED,
  LEFT,
  LAST_SIGNAL
};

/* Drawable internal flags */
enum {
  PGM_DRAWABLE_ROTATED       = (1 << 0),
  PGM_DRAWABLE_SCALED        = (1 << 1),
  PGM_DRAWABLE_PLANE_VECTORS = (1 << 2),
  PGM_DRAWABLE_ORIGIN        = (1 << 3)
};

static GstObjectClass *parent_class = NULL;
static guint pgm_drawable_signals[LAST_SIGNAL] = { 0 };

/* Private methods */

/* Update the transformation matrix for a new drawable position, size or
 * rotation. Called with LOCK. */
static void
update_transformation_matrix (PgmDrawable *drawable)
{
  gfloat x_offset, y_offset;

  /* Get the offset to the center of the drawable */
  x_offset = drawable->x + drawable->width * 0.5f;
  y_offset = drawable->y + drawable->height * 0.5f;

  /* Update the matrix with a translation to the center of the drawable */
  pgm_mat4x4_set_from_scalars (drawable->transformation_matrix,
                               drawable->scale, 0.0f, 0.0f, x_offset,
                               0.0f, drawable->scale, 0.0f, y_offset,
                               0.0f, 0.0f, drawable->scale, drawable->z,
                               0.0f, 0.0f, 0.0f, 1.0f);

  /* Update the rotation from the center */
  pgm_mat4x4_rotate_x (drawable->transformation_matrix, drawable->angle_x);
  pgm_mat4x4_rotate_y (drawable->transformation_matrix, drawable->angle_y);
  pgm_mat4x4_rotate_z (drawable->transformation_matrix, drawable->angle_z);

  /* Reset the origin to the upper-left of the viewport */
  pgm_mat4x4_translate_from_scalars (drawable->transformation_matrix,
                                     -x_offset, -y_offset, -drawable->z);

  /* drawable->origin and drawable->plane_vector_{i,j} are no longer valid */
  drawable->flags &= ~(PGM_DRAWABLE_PLANE_VECTORS | PGM_DRAWABLE_ORIGIN);
}

/* Update plane_vector_i and plane_vector_j if needed. Called with LOCK. */
static void
update_plane_vectors (PgmDrawable *drawable)
{
  PgmVec4 canvas_i, canvas_j;
  PgmVec4 *drawable_i, *drawable_j;

  pgm_vec4_set_from_scalars (&canvas_i, 1.0f, 0.0f, 0.0f, 0.0f);
  pgm_vec4_set_from_scalars (&canvas_j, 0.0f, 1.0f, 0.0f, 0.0f);

  drawable_i = pgm_mat4x4_multiply_vec4 (drawable->transformation_matrix,
                                         &canvas_i);
  pgm_vec3_set_from_scalars (&drawable->plane_vector_i,
                             drawable_i->v[0],
                             drawable_i->v[1],
                             drawable_i->v[2]);

  drawable_j = pgm_mat4x4_multiply_vec4 (drawable->transformation_matrix,
                                         &canvas_j);
  pgm_vec3_set_from_scalars (&drawable->plane_vector_j,
                             drawable_j->v[0],
                             drawable_j->v[1],
                             drawable_j->v[2]);

  drawable->flags |= PGM_DRAWABLE_PLANE_VECTORS;

  pgm_vec4_free (drawable_i);
  pgm_vec4_free (drawable_j);
}

/* Update origin if needed. Called with LOCK. */
static void
update_origin (PgmDrawable *drawable)
{
  PgmVec4 position, *origin4;

  pgm_vec4_set_from_scalars (&position,
                             drawable->x, drawable->y, drawable->z, 1.0f);

  origin4 = pgm_mat4x4_multiply_vec4 (drawable->transformation_matrix,
                                      &position);

  pgm_vec3_set_from_scalars (&drawable->origin,
                             origin4->v[0],
                             origin4->v[1],
                             origin4->v[2]);

  drawable->flags |= PGM_DRAWABLE_ORIGIN;

  pgm_vec4_free (origin4);
}

/* Check whether or not the drag distance has been covered.
 * Called with LOCK. */
static gboolean
is_drag_distance_covered (PgmDrawable *drawable,
                          PgmEventMotion *event)
{
  guchar x_distance =
    (guchar) fabs (drawable->viewport_press_point.v[0] - event->x);
  guchar y_distance =
    (guchar) fabs (drawable->viewport_press_point.v[1] - event->y);

  return (x_distance > drawable->drag_distance ||
          y_distance > drawable->drag_distance);
}

/* Store the intersection between the plane defined by the drawable and a
 * line given by two points. Returns whether or not the intersection belongs
 * to the drawable. */
static gboolean
get_line_intersection (PgmDrawable *drawable,
                       PgmVec3 *intersection,
                       PgmVec3 *p1,
                       PgmVec3 *p2)
{
  gboolean belongs = FALSE;
  PgmVec3 *v;
  gfloat x, y;

  GST_OBJECT_LOCK (drawable);

  if (! (drawable->flags & PGM_DRAWABLE_PLANE_VECTORS))
    update_plane_vectors (drawable);
  if (! (drawable->flags & PGM_DRAWABLE_ORIGIN))
    update_origin (drawable);

  GST_OBJECT_UNLOCK (drawable);

  /* Get intersection with the plane */
  v = pgm_intersection_line_plane (p1, p2, &drawable->origin,
                                   &drawable->plane_vector_i,
                                   &drawable->plane_vector_j);
  if (!v)
    return FALSE;

  /* Store the intersection */
  pgm_vec3_set_from_vec3 (intersection, v);

  /* Test if the intersection belongs to the drawable rectangle */
  pgm_drawable_from_canvas (drawable, &x, &y, v->v[0], v->v[1], v->v[2]);
  belongs = (x >= 0.0f && x <= drawable->width &&
             y >= 0.0f && y <= drawable->height);

  pgm_vec3_free (v);

  return belongs;
}

/* Emit a drawable signal from its index */
static inline gboolean
emit_signal (PgmDrawable *drawable,
             guint signal_index,
             gfloat x,
             gfloat y,
             gfloat z,
             guint time)
{
  gboolean stop_emission;

  g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[signal_index],
                 0, x, y, z, time, &stop_emission);

  return stop_emission;
}

/* Emit the drawable motion signal */
static inline gboolean
emit_motion_signal (PgmDrawable *drawable,
                    gfloat x,
                    gfloat y,
                    gfloat z,
                    guint time)
{
  GST_DEBUG_OBJECT (drawable, "emitting 'motion' signal");

  return emit_signal (drawable, MOTION, x, y, z, time);
}

/* Emit the drawable entered signal */
static inline gboolean
emit_entered_signal (PgmDrawable *drawable,
                     gfloat x,
                     gfloat y,
                     gfloat z,
                     guint time)
{
  GST_DEBUG_OBJECT (drawable, "emitting 'entered' signal");

  return emit_signal (drawable, ENTERED, x, y, z, time);
}

/* Emit the drawable left signal */
static inline void
emit_left_signal (PgmDrawable *drawable,
                  gfloat x,
                  gfloat y,
                  gfloat z,
                  guint time)
{
  GST_DEBUG_OBJECT (drawable, "emitting 'left' signal");

  /* We use g_signal_emit() since left doesn't return anything */
  g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[LEFT],
                 0, x, y, z, time);
}

/* GObject stuff */

G_DEFINE_ABSTRACT_TYPE (PgmDrawable, pgm_drawable, GST_TYPE_OBJECT);

static void
pgm_drawable_dispose (GObject *object)
{
  PgmDrawable *drawable = PGM_DRAWABLE (object);

  pgm_mat4x4_free (drawable->transformation_matrix);

  GST_CALL_PARENT (G_OBJECT_CLASS, dispose, (object));
}

static void
pgm_drawable_class_init (PgmDrawableClass *klass)
{
  GObjectClass *gobject_class = (GObjectClass *) klass;

  parent_class = g_type_class_peek_parent (klass);

  /**
   * PgmDrawable::changed:
   * @drawable: the #PgmDrawable
   * @property: the #PgmDrawableProperty changed in @drawable
   *
   * Will be emitted after @property of @drawable is changed.
   */
  pgm_drawable_signals[CHANGED] =
    g_signal_new ("changed", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_FIRST,
                  G_STRUCT_OFFSET (PgmDrawableClass, changed),
                  NULL, NULL, pgm_marshal_VOID__ENUM, G_TYPE_NONE,
                  1, PGM_TYPE_DRAWABLE_PROPERTY);
  /**
   * PgmDrawable::pressed:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the click on @drawable
   * @y: the y coordinate of the point intersecting the click on @drawable
   * @z: the z coordinate of the point intersecting the click on @drawable
   * @button: the #PgmButtonType
   * @time: the timestamp of the button press event
   * @pressure: the pressure force, set to 0 when not used with a touch screen
   *
   * Will be emitted after @drawable is pressed.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[PRESSED] =
    g_signal_new ("pressed", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, pressed),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT_UINT,
                  G_TYPE_BOOLEAN, 6, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT, G_TYPE_UINT);
  /**
   * PgmDrawable::pressured:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the click on @drawable
   * @y: the y coordinate of the point intersecting the click on @drawable
   * @z: the z coordinate of the point intersecting the click on @drawable
   * @button: the #PgmButtonType
   * @time: the timestamp of the button press event
   * @pressure: the pressure force
   *
   * Will be emitted after the touch pressure on @drawable change.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[PRESSURED] =
    g_signal_new ("pressured", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, pressured),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT_UINT,
                  G_TYPE_BOOLEAN, 6, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT, G_TYPE_UINT);
  /**
   * PgmDrawable::released:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the click on @drawable
   * plane
   * @y: the y coordinate of the point intersecting the click on @drawable
   * plane
   * @z: the z coordinate of the point intersecting the click on @drawable
   * plane
   * @button: the #PgmButtonType
   * @time: the timestamp of the button release event
   *
   * Will be emitted after @drawable is released.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[RELEASED] =
    g_signal_new ("released", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, released),
                  NULL, NULL, pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT,
                  G_TYPE_BOOLEAN, 5, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT);
  /**
   * PgmDrawable::clicked:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the click on @drawable
   * @y: the y coordinate of the point intersecting the click on @drawable
   * @z: the z coordinate of the point intersecting the click on @drawable
   * @button: the #PgmButtonType
   * @time: the timestamp of the button press event
   * @pressure: the pressure force, set to 0 when not used with a touch screen
   *
   * Will be emitted after @drawable is clicked.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[CLICKED] =
    g_signal_new ("clicked", G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, clicked),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT_UINT,
                  G_TYPE_BOOLEAN, 6, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT, G_TYPE_UINT);
  /**
   * PgmDrawable::double-clicked:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the double-click on
   * @drawable
   * @y: the y coordinate of the point intersecting the double-click on
   * @drawable
   * @z: the z coordinate of the point intersecting the double-click on
   * @drawable
   * @button: the #PgmButtonType
   * @time: the timestamp of the button press event
   *
   * Will be emitted after @drawable is clicked.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[DOUBLE_CLICKED] =
    g_signal_new ("double-clicked", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, double_clicked),
                  NULL, NULL, pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT,
                  G_TYPE_BOOLEAN, 5, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT);
  /**
   * PgmDrawable::scrolled:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the scroll on @drawable
   * @y: the y coordinate of the point intersecting the scroll on @drawable
   * @z: the z coordinate of the point intersecting the scroll on @drawable
   * @direction: the #PgmScrollDirection
   * @time: the timestamp of the scroll event
   *
   * Will be emitted after @drawable is scrolled.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[SCROLLED] =
    g_signal_new ("scrolled", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, scrolled),
                  NULL, NULL, pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT,
                  G_TYPE_BOOLEAN, 5, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_SCROLL_DIRECTION, G_TYPE_UINT);
  /**
   * PgmDrawable::drag-begin:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the initial click on
   * @drawable
   * @y: the y coordinate of the point intersecting the initial click on
   * @drawable
   * @z: the z coordinate of the point intersecting the initial click on
   * @drawable
   * @button: the #PgmButtonType
   * @time: the timestamp of the motion event
   * @pressure: the pressure force, set to 0 when not used with a touch screen
   *
   * Will be emitted after @drawable is initially dragged. You can for instance
   * use that signal to change @drawable properties or store the intersecting
   * point for further computations during the "drag-motion" signal emission.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[DRAG_BEGIN] =
    g_signal_new ("drag-begin", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, drag_begin),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT_UINT,
                  G_TYPE_BOOLEAN, 6, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT, G_TYPE_UINT);
  /**
   * PgmDrawable::drag-motion:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the cursor position on
   * the dragged @drawable plane
   * @y: the y coordinate of the point intersecting the cursor position on
   * the dragged @drawable plane
   * @z: the z coordinate of the point intersecting the cursor position on
   * the dragged @drawable plane
   * @button: the #PgmButtonType
   * @time: the timestamp of the motion event
   * @pressure: the pressure force, set to 0 when not used with a touch screen
   *
   * Will be emitted after @drawable is dragged.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[DRAG_MOTION] =
    g_signal_new ("drag-motion", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, drag_motion),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT_UINT,
                  G_TYPE_BOOLEAN, 6, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT, G_TYPE_UINT);
  /**
   * PgmDrawable::drag-end:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the release event on
   * the dragged @drawable plane
   * @y: the y coordinate of the point intersecting the release event on
   * the dragged @drawable plane
   * @z: the z coordinate of the point intersecting the release event on
   * the dragged @drawable plane
   * @button: the #PgmButtonType
   * @time: the timestamp of the motion event
   *
   * Will be emitted after @drawable is released. This signal is emitted on all
   * the drawables which have received a "drag-begin" signal.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[DRAG_END] =
    g_signal_new ("drag-end", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, drag_end),
                  NULL, NULL, pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_ENUM_UINT,
                  G_TYPE_BOOLEAN, 5, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT,
                  PGM_TYPE_BUTTON_TYPE, G_TYPE_UINT);
  /**
   * PgmDrawable::motion:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @y: the y coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @z: the z coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @time: the timestamp of the motion event
   *
   * Will be emitted when the pointer is moved over @drawable.
   *
   * Returns: TRUE to stop other handlers from being invoked for the event.
   * FALSE to propagate the event further.
   */
  pgm_drawable_signals[MOTION] =
    g_signal_new ("motion", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, motion),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_UINT,
                  G_TYPE_BOOLEAN, 4,
                  G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_UINT);
  /**
   * PgmDrawable::entered:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @y: the y coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @z: the z coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @time: the timestamp of the motion event
   *
   * Will be emitted when the pointer enters the @drawable area.
   *
   * Returns: TRUE if drawables under @drawable should not be in "entered" state
   * until the pointer leaves @drawable, FALSE otherwise.
   */
  pgm_drawable_signals[ENTERED] =
    g_signal_new ("entered", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, motion),
                  NULL, NULL,
                  pgm_marshal_BOOLEAN__FLOAT_FLOAT_FLOAT_UINT,
                  G_TYPE_BOOLEAN, 4,
                  G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_UINT);
  /**
   * PgmDrawable::left:
   * @drawable: the #PgmDrawable
   * @x: the x coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @y: the y coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @z: the z coordinate of the point intersecting the cursor position on
   * the @drawable plane
   * @time: the timestamp of the motion event
   *
   * Will be emitted when the pointer leaves the @drawable area.
   *
   */
  pgm_drawable_signals[LEFT] =
    g_signal_new ("left", G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (PgmDrawableClass, motion),
                  NULL, NULL,
                  pgm_marshal_VOID__FLOAT_FLOAT_FLOAT_UINT,
                  G_TYPE_NONE, 4,
                  G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_FLOAT, G_TYPE_UINT);

  gobject_class->dispose = GST_DEBUG_FUNCPTR (pgm_drawable_dispose);

  GST_DEBUG_CATEGORY_INIT (pgm_drawable_debug, "pgm_drawable", 0,
                           "drawable object");
}

static void
pgm_drawable_init (PgmDrawable *drawable)
{
  /* Layering */
  drawable->transformation_matrix = pgm_mat4x4_new_identity ();
  /* drawable->angle_x = 0.0f; */
  /* drawable->angle_y = 0.0f; */
  /* drawable->angle_z = 0.0f; */
  drawable->layer = PGM_DRAWABLE_UNBOUND;
  /* Position */
  /* drawable->x = 0.0f; */
  /* drawable->y = 0.0f; */
  /* drawable->z = 0.0f; */
  /* drawable->orig_x = 0.0f; */
  /* drawable->orig_y = 0.0f; */
  /* Size */
  drawable->width = 1.0f;
  drawable->height = 1.0f;
  drawable->orig_width = 1.0f;
  drawable->orig_height = 1.0f;
  drawable->grid_aligned = FALSE;
  /* Transformation */
  drawable->scale = 1.0f;
  /* Colors */
  drawable->fg_r = 255;
  drawable->fg_g = 255;
  drawable->fg_b = 255;
  drawable->fg_a = 255;
  drawable->bg_r = 255;
  drawable->bg_g = 255;
  drawable->bg_b = 255;
  drawable->bg_a = 255;
  /* Opacity */
  drawable->opacity = 255;
  /* Private data */
  /* pgm_vec3_set_from_scalars (&drawable->origin, 0.0f, 0.0f, 0.0f); */
  pgm_vec3_set_from_scalars (&drawable->plane_vector_i, 1.0f, 0.0f, 0.0f);
  pgm_vec3_set_from_scalars (&drawable->plane_vector_j, 0.0f, 1.0f, 0.0f);
  /* pgm_vec3_set_from_scalars (&drawable->press_point, 0.0f, 0.0f, 0.0f); */
  /* pgm_vec3_set_from_scalars (&drawable->viewport_press_point, 0.0f, 0.0f, 0.0f); */
  /* drawable->pressure = 0; */
  /* drawable->release_time = 0; */
  /* drawable->mouse_over = FALSE; */
  /* drawable->drag_mask = 0; */
  /* drawable->press_mask = 0; */
  /* drawable->release_button = 0; */
  drawable->drag_distance = 5;
  drawable->flags = PGM_DRAWABLE_PLANE_VECTORS | PGM_DRAWABLE_ORIGIN;
}

/* Public methods */

/**
 * pgm_drawable_hide:
 * @drawable: a #PgmDrawable object.
 *
 * Makes @drawable invisible.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_hide (PgmDrawable *drawable)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  GST_OBJECT_FLAG_UNSET (drawable, PGM_DRAWABLE_FLAG_VISIBLE);
  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->hide)
    ret = klass->hide (drawable);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_VISIBILITY);

  return ret;
}

/**
 * pgm_drawable_show:
 * @drawable: a #PgmDrawable object.
 *
 * Makes @drawable visible.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_show (PgmDrawable *drawable)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  GST_OBJECT_FLAG_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE);
  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->show)
    ret = klass->show (drawable);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_VISIBILITY);

  return ret;
}

/**
 * pgm_drawable_is_visible:
 * @drawable: a #PgmDrawable object.
 * @visible: a pointer to a #gboolean where the visible state of the drawable
 * is going to be stored.
 *
 * Retrieves whether @drawable is visible.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_is_visible (PgmDrawable *drawable,
                         gboolean *visible)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (visible != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *visible = GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE);
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_size:
 * @drawable: a #PgmDrawable object.
 * @width: the width of the canvas.
 * @height: the height of the canvas.
 *
 * Defines @drawable size in canvas units to @width x @height.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_size (PgmDrawable *drawable,
                       gfloat width,
                       gfloat height)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;
  PgmCanvas *canvas;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->orig_width = width = MAX (0.0f, width);
  drawable->orig_height = height = MAX (0.0f, height);

  if (drawable->grid_aligned &&
      (canvas = PGM_CANVAS (GST_OBJECT_PARENT (drawable))))
    {
      width = floorf
        (width / canvas->pixel_offset_x) * canvas->pixel_offset_x;
      height = floorf
        (height / canvas->pixel_offset_y) * canvas->pixel_offset_y;
    }

  if (width != drawable->width || height != drawable->height)
    {
      drawable->width = width;
      drawable->height = height;

      /* Update the transformation matrix if the user added transforms */
      if (drawable->flags & (PGM_DRAWABLE_ROTATED | PGM_DRAWABLE_SCALED))
        update_transformation_matrix (drawable);

      GST_OBJECT_UNLOCK (drawable);

      klass = PGM_DRAWABLE_GET_CLASS (drawable);

      if (klass->set_size)
        ret = klass->set_size (drawable, width, height);

      if (drawable->flags & (PGM_DRAWABLE_ROTATED | PGM_DRAWABLE_SCALED))
        _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_TRANSFORMATION_MATRIX);

      _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_SIZE);

      return ret;
    }

  GST_OBJECT_UNLOCK (drawable);

  return ret;
}

/**
 * pgm_drawable_get_size:
 * @drawable: a #PgmDrawable object.
 * @width: a pointer to a #gfloat where the width of the drawable is going to be
 * stored.
 * @height: a pointer to a #gfloat where the height of the drawable is going to
 * be stored.
 *
 * Gets @drawable size in canvas units.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_size (PgmDrawable *drawable,
                       gfloat *width,
                       gfloat *height)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (width != NULL, PGM_ERROR_X);
  g_return_val_if_fail (height != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *width = drawable->orig_width;
  *height = drawable->orig_height;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_position:
 * @drawable: a #PgmDrawable object.
 * @x: the position of @drawable on the x axis.
 * @y: the position of @drawable on the y axis.
 * @z: the position of @drawable on the z axis.
 *
 * Sets position of @drawable in canvas units.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_position (PgmDrawable *drawable,
                           gfloat x,
                           gfloat y,
                           gfloat z)
{
  PgmError ret = PGM_ERROR_OK;
  gboolean reordered = FALSE;
  PgmDrawableClass *klass;
  PgmCanvas *canvas;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->flags &= ~PGM_DRAWABLE_ORIGIN;
  drawable->orig_x = x;
  drawable->orig_y = y;
  canvas = PGM_CANVAS (GST_OBJECT_PARENT (drawable));

  if (drawable->grid_aligned && canvas != NULL)
    {
      x = floorf (x / canvas->pixel_offset_x) * canvas->pixel_offset_x;
      y = floorf (y / canvas->pixel_offset_y) * canvas->pixel_offset_y;
    }

  reordered = (z != drawable->z);

  if (x != drawable->x || y != drawable->y || reordered)
    {
      drawable->x = x;
      drawable->y = y;
      drawable->z = z;

      /* Update the transformation matrix if the user added rotations */
      if (drawable->flags & (PGM_DRAWABLE_ROTATED | PGM_DRAWABLE_SCALED))
        update_transformation_matrix (drawable);

      GST_OBJECT_UNLOCK (drawable);

      klass = PGM_DRAWABLE_GET_CLASS (drawable);

      if (klass->set_position)
        ret = klass->set_position (drawable, x, y, z);

      /* Reorder the drawable inside the canvas in the case that z has changed */
      if (reordered && canvas != NULL)
        _pgm_canvas_update_order (canvas, drawable);

      if (drawable->flags & (PGM_DRAWABLE_ROTATED | PGM_DRAWABLE_SCALED))
        _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_TRANSFORMATION_MATRIX);

      _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_POSITION);

      return ret;
    }

  GST_OBJECT_UNLOCK (drawable);

  return ret;
}

/**
 * pgm_drawable_get_position:
 * @drawable: a #PgmDrawable object.
 * @x: a pointer to a #gfloat where the x position of @drawable is going to
 * be stored.
 * @y: a pointer to a #gfloat where the y position of @drawable is going to
 * be stored.
 * @z: a pointer to a #gfloat where the z position of @drawable is going to
 * be stored.
 *
 * Gets @drawable position in canvas units.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_position (PgmDrawable *drawable,
                           gfloat *x,
                           gfloat *y,
                           gfloat *z)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (x != NULL, PGM_ERROR_X);
  g_return_val_if_fail (y != NULL, PGM_ERROR_X);
  g_return_val_if_fail (z != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *x = drawable->orig_x;
  *y = drawable->orig_y;
  *z = drawable->z;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_rotation_x:
 * @drawable: a #PgmDrawable object.
 * @angle: an angle of rotation, in radians.
 *
 * Defines the rotation to apply to @drawable on the x axis. The rotation point
 * is the center of @drawable and the axis is its normal.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_rotation_x (PgmDrawable *drawable,
                             gfloat angle)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (angle == drawable->angle_x)
    {
      GST_OBJECT_UNLOCK (drawable);
      return PGM_ERROR_OK;
    }

  drawable->angle_x = angle;
  update_transformation_matrix (drawable);

  /* Update flags to avoid further unneeded computations */
  if (angle != 0.0f)
    drawable->flags |= PGM_DRAWABLE_ROTATED;
  else if (drawable->angle_y == 0.0f && drawable->angle_z == 0.0f)
    drawable->flags &= ~PGM_DRAWABLE_ROTATED;

  GST_OBJECT_UNLOCK (drawable);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_TRANSFORMATION_MATRIX);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_get_rotation_x:
 * @drawable: a #PgmDrawable object.
 * @angle: a #gfloat pointer where the x rotation angle is going to be stored.
 *
 * Retrieves in @angle the current x rotation angle applied to @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_rotation_x (PgmDrawable *drawable,
                             gfloat *angle)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (angle != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *angle = drawable->angle_x;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_rotation_y:
 * @drawable: a #PgmDrawable object.
 * @angle: an angle of rotation, in radians.
 *
 * Defines the rotation to apply to @drawable on the y axis. The rotation point
 * is the center of @drawable and the axis is its normal.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_rotation_y (PgmDrawable *drawable,
                             gfloat angle)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (angle == drawable->angle_y)
    {
      GST_OBJECT_UNLOCK (drawable);
      return PGM_ERROR_OK;
    }

  drawable->angle_y = angle;
  update_transformation_matrix (drawable);

  /* Update flags to avoid further unneeded computations */
  if (angle != 0.0f)
    drawable->flags |= PGM_DRAWABLE_ROTATED;
  else if (drawable->angle_x == 0.0f && drawable->angle_z == 0.0f)
    drawable->flags &= ~PGM_DRAWABLE_ROTATED;

  GST_OBJECT_UNLOCK (drawable);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_TRANSFORMATION_MATRIX);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_get_rotation_y:
 * @drawable: a #PgmDrawable object.
 * @angle: a #gfloat pointer where the y rotation angle is going to be stored.
 *
 * Retrieves in @angle the current y rotation angle applied to @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_rotation_y (PgmDrawable *drawable,
                             gfloat *angle)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (angle != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *angle = drawable->angle_y;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_rotation_z:
 * @drawable: a #PgmDrawable object.
 * @angle: an angle of rotation, in radians.
 *
 * Defines the rotation to apply to @drawable on the z axis. The rotation point
 * is the center of @drawable and the axis is its normal.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_rotation_z (PgmDrawable *drawable,
                             gfloat angle)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (angle == drawable->angle_z)
    {
      GST_OBJECT_UNLOCK (drawable);
      return PGM_ERROR_OK;
    }

  drawable->angle_z = angle;
  update_transformation_matrix (drawable);

  /* Update flags to avoid further unneeded computations */
  if (angle != 0.0f)
    drawable->flags |= PGM_DRAWABLE_ROTATED;
  else if (drawable->angle_x == 0.0f && drawable->angle_y == 0.0f)
    drawable->flags &= ~PGM_DRAWABLE_ROTATED;

  GST_OBJECT_UNLOCK (drawable);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_TRANSFORMATION_MATRIX);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_get_rotation_z:
 * @drawable: a #PgmDrawable object.
 * @angle: a #gfloat pointer where the z rotation angle is going to be stored.
 *
 * Retrieves in @angle the current z rotation angle applied to @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_rotation_z (PgmDrawable *drawable,
                             gfloat *angle)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (angle != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *angle = drawable->angle_z;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_scale:
 * @drawable: a #PgmDrawable object.
 * @scale: a scale factor.
 *
 * Defines the scale to apply to @drawable. The scaling point is the center of
 * @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_scale (PgmDrawable *drawable,
                        gfloat scale)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (scale == drawable->scale)
    {
      GST_OBJECT_UNLOCK (drawable);
      return PGM_ERROR_OK;
    }

  drawable->scale = scale;
  update_transformation_matrix (drawable);

  /* Update flags to avoid further unneeded computations */
  if (scale != 1.0f)
    drawable->flags |= PGM_DRAWABLE_SCALED;
  else
    drawable->flags &= ~PGM_DRAWABLE_SCALED;

  GST_OBJECT_UNLOCK (drawable);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_TRANSFORMATION_MATRIX);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_get_scale:
 * @drawable: a #PgmDrawable object.
 * @scale: a #gfloat pointer where the scale factor is going to be stored.
 *
 * Retrieves in @scale the current scale factor applied to @drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_scale (PgmDrawable *drawable,
                        gfloat *scale)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (scale != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *scale = drawable->scale;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_fg_color:
 * @drawable: a #PgmDrawable object.
 * @r: the red foreground color.
 * @g: the green foreground color.
 * @b: the blue foreground color.
 * @a: the alpha foreground color.
 *
 * Set the color used for drawing operations on @drawable to (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_fg_color (PgmDrawable *drawable,
                           guchar r,
                           guchar g,
                           guchar b,
                           guchar a)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->fg_r = r;
  drawable->fg_g = g;
  drawable->fg_b = b;
  drawable->fg_a = a;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_fg_color)
    ret = klass->set_fg_color (drawable, r, g, b, a);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_FG_COLOR);

  return ret;
}

/**
 * pgm_drawable_get_fg_color:
 * @drawable: a #PgmDrawable object.
 * @r: a pointer to a #guchar where the red foreground color is going to be
 * stored.
 * @g: a pointer to a #guchar where the green foreground color is going to be
 * stored.
 * @b: a pointer to a #guchar where the blue foreground color is going to be
 * stored.
 * @a: a pointer to a #guchar where the alpha foreground color is going to be
 * stored.
 *
 * Retrieves the color used for drawing operations on @drawable in
 * (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_fg_color (PgmDrawable *drawable,
                           guchar *r,
                           guchar *g,
                           guchar *b,
                           guchar *a)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (r != NULL, PGM_ERROR_X);
  g_return_val_if_fail (g != NULL, PGM_ERROR_X);
  g_return_val_if_fail (b != NULL, PGM_ERROR_X);
  g_return_val_if_fail (a != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *r = drawable->fg_r;
  *g = drawable->fg_g;
  *b = drawable->fg_b;
  *a = drawable->fg_a;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_bg_color:
 * @drawable: a #PgmDrawable object.
 * @r: the red background color.
 * @g: the green background color.
 * @b: the blue background color.
 * @a: the alpha background color.
 *
 * Sets the color used to fill the background of @drawable to (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_bg_color (PgmDrawable *drawable,
                           guchar r,
                           guchar g,
                           guchar b,
                           guchar a)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->bg_r = r;
  drawable->bg_g = g;
  drawable->bg_b = b;
  drawable->bg_a = a;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_bg_color)
    ret = klass->set_bg_color (drawable, r, g, b, a);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_BG_COLOR);

  return ret;
}

/**
 * pgm_drawable_get_bg_color:
 * @drawable: A #PgmDrawable object.
 * @r: a pointer to a #guchar where the red background color is going to be
 * stored.
 * @g: a pointer to a #guchar where the green background color is going to be
 * stored.
 * @b: a pointer to a #guchar where the blue background color is going to be
 * stored.
 * @a: a pointer to a #guchar where the alpha background color is going to be
 * stored.
 *
 * Retrieves the color used to fill the background of @drawable in
 * (@r,@g,@b,@a).
 *
 * MT safe.
 *
 * Returns: A #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_bg_color (PgmDrawable *drawable,
                           guchar *r,
                           guchar *g,
                           guchar *b,
                           guchar *a)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (r != NULL, PGM_ERROR_X);
  g_return_val_if_fail (g != NULL, PGM_ERROR_X);
  g_return_val_if_fail (b != NULL, PGM_ERROR_X);
  g_return_val_if_fail (a != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  *r = drawable->bg_r;
  *g = drawable->bg_g;
  *b = drawable->bg_b;
  *a = drawable->bg_a;

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_opacity:
 * @drawable: a #PgmDrawable object.
 * @opacity: the opacity of @drawable.
 *
 * Defines the opacity of @drawable to @opacity.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_opacity (PgmDrawable *drawable,
                          guchar opacity)
{
  PgmError ret = PGM_ERROR_OK;
  PgmDrawableClass *klass;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  drawable->opacity = opacity;

  GST_OBJECT_UNLOCK (drawable);

  klass = PGM_DRAWABLE_GET_CLASS (drawable);

  if (klass->set_opacity)
    ret = klass->set_opacity (drawable, opacity);

  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_OPACITY);

  return ret;
}

/**
 * pgm_drawable_get_opacity:
 * @drawable: a #PgmDrawable object.
 * @opacity: a pointer to a #guchar where the opacity is going to be stored.
 *
 * Retrieves the opacity of @drawable in @opacity.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_opacity (PgmDrawable *drawable,
                          guchar *opacity)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (opacity != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *opacity = drawable->opacity;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_grid_aligned:
 * @drawable: a #PgmDrawable object.
 * @aligned: the grid aligned boolean. %FALSE by default.
 *
 * Sets whether or not @drawable must be automatically aligned to the pixel
 * boundaries in the viewport.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_grid_aligned (PgmDrawable *drawable,
                               gboolean aligned)
{
  gfloat x, y, z, w, h;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (aligned == drawable->grid_aligned)
    {
      GST_OBJECT_UNLOCK (drawable);
      return PGM_ERROR_OK;
    }

  drawable->grid_aligned = aligned;

  x = drawable->orig_x;
  y = drawable->orig_y;
  z = drawable->z;
  w = drawable->orig_width;
  h = drawable->orig_height;

  GST_OBJECT_UNLOCK (drawable);

  pgm_drawable_set_position (drawable, x, y, z);
  pgm_drawable_set_size (drawable, w, h);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_get_grid_aligned:
 * @drawable: a #PgmDrawable object.
 * @aligned: a pointer to a #gboolean where the grid aligned boolean is going to
 * be stored.
 *
 * Gets whether or not @drawable must be automatically aligned to the pixel
 * boundaries in the viewport.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_grid_aligned (PgmDrawable *drawable,
                               gboolean *aligned)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (aligned != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *aligned = drawable->grid_aligned;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_set_drag_distance:
 * @drawable: a #PgmDrawable object.
 * @distance: the drag distance of @drawable. 5 by default.
 *
 * Defines the drag distance of @drawable to @distance. The drag distance is the
 * distance in pixels the user has to cover with a mouse button pressed to
 * trigger the <link linkend="PgmDrawable-drag-begin">'drag-begin'</link> signal.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_set_drag_distance (PgmDrawable *drawable,
                                guchar distance)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  drawable->drag_distance = distance;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_get_drag_distance:
 * @drawable: a #PgmDrawable object.
 * @distance: a pointer to a #guchar where the drag distance is going to be
 * stored.
 *
 * Retrieves the drag distance of @drawable in @distance.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_get_drag_distance (PgmDrawable *drawable,
                                guchar *distance)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (distance != NULL, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);
  *distance = drawable->drag_distance;
  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_regenerate:
 * @drawable: a #PgmDrawable object.
 *
 * Forces a regeneration of the drawable.
 *
 * MT safe.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_regenerate (PgmDrawable *drawable)
{
  _pgm_drawable_emit_changed (drawable, PGM_DRAWABLE_REGENERATE);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_from_canvas:
 * @drawable: a #PgmDrawable object.
 * @x_drawable: where to return the x coordinate in the @drawable plane.
 * @y_drawable: where to return the y coordinate in the @drawable plane.
 * @x_canvas: x coordinate in the canvas space.
 * @y_canvas: y coordinate in the canvas space.
 * @z_canvas: z coordinate in the canvas space.
 *
 * Converts a position in the canvas into its orthogonal projection in the plane
 * of @drawable, with the (0,0) coordinate in that plane being the upper left
 * corner of @drawable.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_from_canvas (PgmDrawable *drawable,
                          gfloat *x_drawable,
                          gfloat *y_drawable,
                          gfloat x_canvas,
                          gfloat y_canvas,
                          gfloat z_canvas)
{
  PgmVec3 v;

  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (NULL != x_drawable, PGM_ERROR_X);
  g_return_val_if_fail (NULL != y_drawable, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (! (drawable->flags & PGM_DRAWABLE_PLANE_VECTORS))
    update_plane_vectors (drawable);
  if (! (drawable->flags & PGM_DRAWABLE_ORIGIN))
    update_origin (drawable);

  /* v is the vector going from the top left corner of the drawable to the
   * considered point. We project it on the unit vectors along the top and left
   * sides of the drawable. This is equivalent (but faster) to applying the
   * inverse of drawable->transformation_matrix to the vector
   * (x_canvas, y_canvas, z_canvas, 1). */

  pgm_vec3_set_from_scalars (&v,
                             x_canvas - drawable->origin.v[0],
                             y_canvas - drawable->origin.v[1],
                             z_canvas - drawable->origin.v[2]);

  *x_drawable = pgm_vec3_dot_product (&v, &drawable->plane_vector_i);
  *y_drawable = pgm_vec3_dot_product (&v, &drawable->plane_vector_j);

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}

/**
 * pgm_drawable_to_canvas:
 * @drawable: a #PgmDrawable object.
 * @x_canvas: where to return the x coordinate in the canvas space.
 * @y_canvas: where to return the y coordinate in the canvas space.
 * @z_canvas: where to return the z coordinate in the canvas space.
 * @x_drawable: x coordinate in the @drawable plane.
 * @y_drawable: y coordinate in the @drawable plane.
 *
 * Converts a position in the plane of @drawable into the canvas space.
 *
 * Returns: a #PgmError indicating success/failure.
 */
PgmError
pgm_drawable_to_canvas (PgmDrawable *drawable,
                        gfloat *x_canvas,
                        gfloat *y_canvas,
                        gfloat *z_canvas,
                        gfloat x_drawable,
                        gfloat y_drawable)
{
  g_return_val_if_fail (PGM_IS_DRAWABLE (drawable), PGM_ERROR_X);
  g_return_val_if_fail (NULL != x_canvas, PGM_ERROR_X);
  g_return_val_if_fail (NULL != y_canvas, PGM_ERROR_X);
  g_return_val_if_fail (NULL != z_canvas, PGM_ERROR_X);

  GST_OBJECT_LOCK (drawable);

  if (! (drawable->flags & PGM_DRAWABLE_PLANE_VECTORS))
    update_plane_vectors (drawable);
  if (! (drawable->flags & PGM_DRAWABLE_ORIGIN))
    update_origin (drawable);

  /* The resulting (x_canvas, y_canvas, z_canvas) vector is:
   * origin + x_drawable * plane_vector_i + y_drawable * plane_vector_j
   * This is equivalent to applying drawable->transformation_matrix to the
   * (x_drawable, y_drawable, 0, 1) vector. */
  *x_canvas = drawable->origin.v[0]
              + x_drawable * drawable->plane_vector_i.v[0]
              + y_drawable * drawable->plane_vector_j.v[0];
  *y_canvas = drawable->origin.v[1]
              + x_drawable * drawable->plane_vector_i.v[1]
              + y_drawable * drawable->plane_vector_j.v[1];
  *z_canvas = drawable->origin.v[2]
              + x_drawable * drawable->plane_vector_i.v[2]
              + y_drawable * drawable->plane_vector_j.v[2];

  GST_OBJECT_UNLOCK (drawable);

  return PGM_ERROR_OK;
}


/* Protected methods */

/**
 * _pgm_drawable_emit_changed:
 * @drawable: the #PgmDrawable object.
 * @property: the #PgmDrawablePoroperty that has changed.
 *
 * This function is only used internally by #PgmDrawable and its subclasses to
 * emit the 'changed' signal on @drawable.
 *
 * MT safe.
 */
void
_pgm_drawable_emit_changed (PgmDrawable *drawable,
                            PgmDrawableProperty property)
{
  g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[CHANGED], 0,
                 property);
}

/**
 * _pgm_drawable_do_press_event:
 * @drawable: the #PgmDrawable object.
 * @event: the #PgmEventButton.
 * @p1: the 1st line point in canvas coordinates.
 * @p2: the 2nd line point in canvas coordinates.
 * @emission_mask: the mask to inhibit signal emission.
 *
 * This function is only used internally in #PgmViewport to propagate mouse
 * press events to drawables so that they can properly handle and emit the
 * corresponding picking signals.
 *
 * MT safe.
 */
void
_pgm_drawable_do_press_event (PgmDrawable *drawable,
                              PgmEventButton *event,
                              PgmVec3 *p1,
                              PgmVec3 *p2,
                              guint16 *emission_mask)
{
  PgmVec3 intersect;

  /* Do not process invisible drawables */
  if (!GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE)
      || drawable->opacity == 0 || drawable->fg_a == 0)
    return;

  /* Update properties and emit the signal if the cursor is over the drawable */
  if (get_line_intersection (drawable, &intersect, p1, p2))
    {
      drawable->press_mask |= event->button;
      pgm_vec3_set_from_vec3 (&drawable->press_point, &intersect);
      pgm_vec3_set_from_scalars (&drawable->viewport_press_point,
                                 event->x, event->y, 0.0f);

      if (*emission_mask & PGM_DRAWABLE_PICKING_PRESSED)
        {
          gboolean stop_emission;

          GST_DEBUG_OBJECT (drawable, "emitting 'pressed' signal");
          g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[PRESSED], 0,
                         intersect.v[0], intersect.v[1], intersect.v[2],
                         event->button, event->time, event->pressure,
                         &stop_emission);
          if (stop_emission)
            *emission_mask &= ~PGM_DRAWABLE_PICKING_PRESSED;
        }
    }
}

/**
 * _pgm_drawable_do_pressure_event:
 * @drawable: the #PgmDrawable object.
 * @event: the #PgmEventButton.
 * @p1: the 1st line point in canvas coordinates.
 * @p2: the 2nd line point in canvas coordinates.
 * @emission_mask: the mask to inhibit signal emission.
 *
 * This function is only used internally in #PgmViewport to propagate mouse
 * pressure change events to drawables so that they can properly handle and
 * emit the corresponding picking signals.
 *
 * MT safe.
 */
void
_pgm_drawable_do_pressure_event (PgmDrawable *drawable,
                                 PgmEventButton *event,
                                 PgmVec3 *p1,
                                 PgmVec3 *p2,
                                 guint16 *emission_mask)
{
  /* Do not process invisible drawables */
  if ((!GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE)
       || drawable->opacity == 0 || drawable->fg_a == 0)
      && !drawable->drag_mask)
    return;

  /* First check if the button is pressed before going to further computations */
  if (drawable->press_mask & event->button)
    {
      gboolean stop_emission;
      PgmVec3 intersect;

      /* Emit the pressure signal */
      if (*emission_mask & PGM_DRAWABLE_PICKING_PRESSURED)
        {
          get_line_intersection (drawable, &intersect, p1, p2);

          GST_DEBUG_OBJECT (drawable, "emitting 'pressured' signal");
          g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[PRESSURED],
                         0, intersect.v[0], intersect.v[1], intersect.v[2],
                         event->button, event->time, event->pressure,
                         &stop_emission);
          if (stop_emission)
            *emission_mask &= ~PGM_DRAWABLE_PICKING_PRESSURED;
        }

      drawable->pressure = event->pressure;
    }
}

/**
 * _pgm_drawable_do_release_event:
 * @drawable: the #PgmDrawable object.
 * @event: the #PgmEventButton.
 * @p1: the 1st line point in canvas coordinates.
 * @p2: the 2nd line point in canvas coordinates.
 * @emission_mask: the mask to inhibit signal emission.
 *
 * This function is only used internally in #PgmViewport to propagate mouse
 * release events to drawables so that they can properly handle and emit the
 * corresponding picking signals.
 *
 * MT safe.
 */
void
_pgm_drawable_do_release_event (PgmDrawable *drawable,
                                PgmEventButton *event,
                                PgmVec3 *p1,
                                PgmVec3 *p2,
                                guint16 *emission_mask)
{
  /* Do not process invisible drawables */
  if ((!GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE)
       || drawable->opacity == 0 || drawable->fg_a == 0)
      && !drawable->drag_mask)
    return;

  /* First check if the button is pressed before going to further computations */
  if (drawable->press_mask & event->button)
    {
      gboolean cached = FALSE;
      gboolean picked = FALSE;
      gboolean stop_emission;
      PgmVec3 intersect;

      /* Emit the release signal */
      if (*emission_mask & PGM_DRAWABLE_PICKING_RELEASED)
        {
          picked = get_line_intersection (drawable, &intersect, p1, p2);
          cached = TRUE;

          GST_DEBUG_OBJECT (drawable, "emitting 'released' signal");
          g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[RELEASED], 0,
                         intersect.v[0], intersect.v[1], intersect.v[2],
                         event->button, event->time, &stop_emission);
          if (stop_emission)
            *emission_mask &= ~PGM_DRAWABLE_PICKING_RELEASED;
        }

      /* Emit a click if the cursor is over the drawable. Take care of not
       * emitting it if there's an on-going drag. */
      if (!(drawable->drag_mask & event->button)
          && *emission_mask & PGM_DRAWABLE_PICKING_CLICKED)
        {
          if (!cached)
            {
              picked = get_line_intersection (drawable, &intersect, p1, p2);
              cached = TRUE;
            }

          if (picked)
            {
              GST_DEBUG_OBJECT (drawable, "emitting 'clicked' signal");
              g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[CLICKED],
                             0, intersect.v[0], intersect.v[1], intersect.v[2],
                             event->button, event->time, drawable->pressure,
                             &stop_emission);
              if (stop_emission)
                *emission_mask &= ~PGM_DRAWABLE_PICKING_CLICKED;
            }
        }

      /* Emit a double-click if the cursor is over the drawable, the last
       * released button is the same and the last click happened recently */
      if (*emission_mask & PGM_DRAWABLE_PICKING_DOUBLE_CLICKED)
        {
          if (!cached)
            {
              picked = get_line_intersection (drawable, &intersect, p1, p2);
              cached = TRUE;
            }

          if (picked)
            {
              if (drawable->release_button & event->button
                  && event->time - drawable->release_time < DOUBLE_CLICK_DELAY)
                {
                  GST_DEBUG_OBJECT (drawable,
                                    "emitting 'double-clicked' signal");
                  g_signal_emit (G_OBJECT (drawable),
                                 pgm_drawable_signals[DOUBLE_CLICKED], 0,
                                 intersect.v[0], intersect.v[1], intersect.v[2],
                                 event->button, event->time, &stop_emission);
                  if (stop_emission)
                    *emission_mask &= ~PGM_DRAWABLE_PICKING_DOUBLE_CLICKED;

                  drawable->release_time = 0;
                  drawable->release_button = 0;
                }
              else
                {
                  drawable->release_time = event->time;
                  drawable->release_button |= event->button;
                }
            }
        }

      /* Emit a drag end if there was an on-going dragging */
      if (drawable->drag_mask & event->button)
        {
          if (*emission_mask & PGM_DRAWABLE_PICKING_DRAG_END)
            {
              if (!cached)
                get_line_intersection (drawable, &intersect, p1, p2);

              GST_DEBUG_OBJECT (drawable, "emitting 'drag-end' signal");
              g_signal_emit (G_OBJECT (drawable),
                             pgm_drawable_signals[DRAG_END], 0,
                             intersect.v[0], intersect.v[1], intersect.v[2],
                             event->button, event->time, &stop_emission);
              if (stop_emission)
                *emission_mask &= ~PGM_DRAWABLE_PICKING_DRAG_END;
            }
          drawable->drag_mask &= ~event->button;
        }

      drawable->press_mask &= ~event->button;
    }
}

/**
 * _pgm_drawable_do_motion_event:
 * @drawable: the #PgmDrawable object.
 * @event: the #PgmEventButton.
 * @p1: the 1st line point in canvas coordinates.
 * @p2: the 2nd line point in canvas coordinates.
 * @emission_mask: the mask to inhibit signal emission.
 *
 * This function is only used internally in #PgmViewport to propagate mouse
 * motion events to drawables so that they can properly handle and emit the
 * corresponding picking signals.
 *
 * MT safe.
 */
void
_pgm_drawable_do_motion_event (PgmDrawable *drawable,
                               PgmEventMotion *event,
                               PgmVec3 *p1,
                               PgmVec3 *p2,
                               guint16 *emission_mask)
{
  gboolean stop_emission;

  /* Do not process invisible drawables */
  if ((!GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE)
       || drawable->opacity == 0 || drawable->fg_a == 0)
      && !drawable->mouse_over)
    return;

  /* Avoid further computations if there's no button pressed */
  if (drawable->press_mask)
    {
      guint16 button;

      /* Check for pressed buttons */
      for (button = 1; button <= PGM_BUTTON_RIGHT; button <<= 1)
        {
          if (drawable->press_mask & button)
            {
              /* Emit the drag begin signal if there's no on-going drag */
              if (!(drawable->drag_mask & button)
                  && is_drag_distance_covered (drawable, event))
                {
                  drawable->drag_mask |= button;

                  if (*emission_mask & PGM_DRAWABLE_PICKING_DRAG_BEGIN)
                    {
                      GST_DEBUG_OBJECT (drawable,
                                        "emitting 'drag-begin' signal");
                      g_signal_emit (G_OBJECT (drawable),
                                     pgm_drawable_signals[DRAG_BEGIN],
                                     0, drawable->press_point.v[0],
                                     drawable->press_point.v[1],
                                     drawable->press_point.v[2],
                                     button, event->time, event->pressure,
                                     &stop_emission);
                      if (stop_emission)
                        *emission_mask &= ~PGM_DRAWABLE_PICKING_DRAG_BEGIN;
                    }
                }

              /* Emit the drag motion signal */
              if ((*emission_mask & PGM_DRAWABLE_PICKING_DRAG_MOTION)
                  && (drawable->drag_mask & button))
                {
                  PgmVec3 intersect;

                  get_line_intersection (drawable, &intersect, p1, p2);

                  GST_DEBUG_OBJECT (drawable, "emitting 'drag-motion' signal");
                  g_signal_emit (G_OBJECT (drawable),
                                 pgm_drawable_signals[DRAG_MOTION],
                                 0, intersect.v[0], intersect.v[1],
                                 intersect.v[2], button, event->time,
                                 event->pressure, &stop_emission);
                  if (stop_emission)
                    *emission_mask &= ~PGM_DRAWABLE_PICKING_DRAG_MOTION;
                }
            }
        }
    }
  else if (*emission_mask & (PGM_DRAWABLE_PICKING_MOTION
                             | PGM_DRAWABLE_PICKING_ENTERED))
    {
      PgmDrawable *emission_stopper;
      PgmCanvas *canvas;
      gboolean on_drawable;
      PgmVec3 intersect;
      gfloat x, y, z;

      /* Get the intersection and the emission stopper */
      canvas = PGM_CANVAS (GST_OBJECT_PARENT (drawable));
      on_drawable = get_line_intersection (drawable, &intersect, p1, p2);
      emission_stopper = _pgm_canvas_get_entered_emission_stopper (canvas);

      /* Store the intersection in local variables */
      x = intersect.v[0];
      y = intersect.v[1];
      z = intersect.v[2];

      /* If we're leaving a drawable */
      if (!on_drawable && drawable->mouse_over)
        {
          emit_left_signal (drawable, x, y, z, event->time);
          drawable->mouse_over = FALSE;
          if (emission_stopper == drawable)
            {
              _pgm_canvas_set_entered_emission_stopper (canvas, NULL);
              *emission_mask |= PGM_DRAWABLE_PICKING_ENTERED;
            }
        }
      else if (on_drawable)
        {
          /* If drawable is the emission stopper */
          if (emission_stopper == drawable)
            *emission_mask &= ~PGM_DRAWABLE_PICKING_ENTERED;

          /* If we're entering a drawable */
          else if ((*emission_mask & PGM_DRAWABLE_PICKING_ENTERED)
                   && !drawable->mouse_over)
            {
              drawable->mouse_over = TRUE;

              /* If the entered handler returns TRUE, drawables below it can't
               * be in "entered" state */
              if (emit_entered_signal (drawable, x, y, z, event->time))
                {
                  _pgm_canvas_set_entered_emission_stopper (canvas, drawable);
                  *emission_mask &= ~PGM_DRAWABLE_PICKING_ENTERED;

                  /* If there was an emission stopper *below* us, emit left on
                   * the old emission stopper */
                  if (NULL != emission_stopper)
                    {
                      emit_left_signal (emission_stopper,
                                        x, y, z, event->time);
                      emission_stopper->mouse_over = FALSE;
                    }
                }
            }

          /* The motion signal emission is here since we want it to be emitted
           * after the entered signal */
          if (*emission_mask & PGM_DRAWABLE_PICKING_MOTION)
            {
              if (emit_motion_signal (drawable, x, y, z, event->time))
                *emission_mask &= ~PGM_DRAWABLE_PICKING_MOTION;
            }
        }
    }
}

/**
 * _pgm_drawable_do_scroll_event:
 * @drawable: the #PgmDrawable object.
 * @event: the #PgmEventButton.
 * @p1: the 1st line point in canvas coordinates.
 * @p2: the 2nd line point in canvas coordinates.
 * @emission_mask: the mask to inhibit signal emission.
 *
 * This function is only used internally in #PgmViewport to propagate mouse
 * scroll events to drawables so that they can properly handle and emit the
 * corresponding picking signals.
 *
 * MT safe.
 */
void
_pgm_drawable_do_scroll_event (PgmDrawable *drawable,
                               PgmEventScroll *event,
                               PgmVec3 *p1,
                               PgmVec3 *p2,
                               guint16 *emission_mask)
{
  PgmVec3 intersect;

  /* Do not process invisible drawables */
  if (!GST_OBJECT_FLAG_IS_SET (drawable, PGM_DRAWABLE_FLAG_VISIBLE)
      || drawable->opacity == 0 || drawable->fg_a == 0)
    return;

  /* Update properties and emit the signal if the cursor is over the drawable */
  if (get_line_intersection (drawable, &intersect, p1, p2))
    {
      pgm_vec3_set_from_vec3 (&drawable->press_point, &intersect);

      if (*emission_mask & PGM_DRAWABLE_PICKING_SCROLLED)
        {
          gboolean stop_emission;

          GST_DEBUG_OBJECT (drawable, "emitting 'scrolled' signal");
          g_signal_emit (G_OBJECT (drawable), pgm_drawable_signals[SCROLLED], 0,
                         intersect.v[0], intersect.v[1], intersect.v[2],
                         event->direction, event->time, &stop_emission);
          if (stop_emission)
            *emission_mask &= ~PGM_DRAWABLE_PICKING_SCROLLED;
        }
    }
}
